#include "IrrlichtDrawDevice.h"
#include "LayerManagerInfo.h"

extern void message_log(const char* format, ...);
extern void error_log(const char *format, ...);

using namespace irr;
using namespace core;
using namespace video;
using namespace scene;
using namespace io;
using namespace gui;

/**
 * RXgN^
 */
IrrlichtDrawDevice::IrrlichtDrawDevice(iTJSDispatch2 *objthis, int width, int height)
	: IrrlichtBase(objthis), width(width), height(height), destWidth(0), destHeight(0), zoomMode(true), defaultVisible(true)
{
	// IrrlichtIʃTCY
	screenWidth = width;
	screenHeight = height;
	screenRect = rect<s32>(0,0,screenWidth,screenHeight);
}

/**
 * fXgN^
 */
IrrlichtDrawDevice::~IrrlichtDrawDevice()
{
	stop();
	detach();
}

/**
 * t@Ng
 */
tjs_error
IrrlichtDrawDevice::Factory(IrrlichtDrawDevice **obj, tjs_int numparams, tTJSVariant **param, iTJSDispatch2 *objthis)
{
	if (numparams < 2) {
		return TJS_E_BADPARAMCOUNT;
	}
	*obj = new IrrlichtDrawDevice(objthis, (tjs_int)*param[0], (tjs_int)*param[1]);
	return TJS_S_OK;
}

// -----------------------------------------------------------------------
// Continuous
// -----------------------------------------------------------------------

/**
 * Continuous R[obN
 * ggɂȂƂɏɌĂ΂
 */
void TJS_INTF_METHOD
IrrlichtDrawDevice::OnContinuousCallback(tjs_uint64 tick)
{
	Window->RequestUpdate();
}

void
IrrlichtDrawDevice::onAttach()
{
	if (device) {
		Window->NotifySrcResize(); // ĂԂƂ GetSrcSize(), SetDestRectangle() ̌ĂѕԂ
		// }l[Wɑ΂eNX`̊蓖
		IVideoDriver *driver = device->getVideoDriver();
		if (driver) {
			for (std::vector<iTVPLayerManager *>::iterator i = Managers.begin(); i != Managers.end(); i++) {
				iTVPLayerManager *manager = *i;
				LayerManagerInfo *info = (LayerManagerInfo*)manager->GetDrawDeviceData();
				if (info != NULL) {
					info->alloc(manager, driver);
				}
			}
		}
	}
}

void
IrrlichtDrawDevice::onDetach()
{
	for (std::vector<iTVPLayerManager *>::iterator i = Managers.begin(); i != Managers.end(); i++) {
		LayerManagerInfo *info = (LayerManagerInfo*)(*i)->GetDrawDeviceData();
		if (info) {
			info->free();
		}
	}
}

/**
 * DeviceIrrlicht̍W̕ϊs
 * @param		x		Xʒu
 * @param		y		Yʒu
 * @note		x, y  DestRect (0,0) _ƂWƂēnƌȂ
 */
void
IrrlichtDrawDevice::transformToIrrlicht(tjs_int &x, tjs_int &y)
{
	x = screenWidth  ? (x * destWidth  / screenWidth) : 0;
	y = screenHeight ? (y * destHeight / screenHeight) : 0;
}

/** IrrlichtDevice̍W̕ϊs
 * @param		x		Xʒu
 * @param		y		Yʒu
 * @note		x, y  C (0,0) _ƂWƂēnƌȂ
 */
void
IrrlichtDrawDevice::transformFromIrrlicht(tjs_int &x, tjs_int &y)
{
	x = destWidth  ? (x * screenWidth  / destWidth) : 0;
	y = destHeight ? (y * screenHeight / destHeight) : 0;
}

/**
 * DevicevC}C̍W̕ϊs
 * @param		x		Xʒu
 * @param		y		Yʒu
 * @note		x, y  DestRect (0,0) _ƂWƂēnƌȂ
 */
void
IrrlichtDrawDevice::transformToManager(iTVPLayerManager * manager, tjs_int &x, tjs_int &y)
{
	// vC}C}l[W̃vC}C̃TCY𓾂
	tjs_int pl_w, pl_h;
	manager->GetPrimaryLayerSize(pl_w, pl_h);
	x = destWidth  ? (x * pl_w / destWidth) : 0;
	y = destHeight ? (y * pl_h / destHeight) : 0;
}

/** vC}CDevice̍W̕ϊs
 * @param		x		Xʒu
 * @param		y		Yʒu
 * @note		x, y  C (0,0) _ƂWƂēnƌȂ
 */
void
IrrlichtDrawDevice::transformFromManager(iTVPLayerManager * manager, tjs_int &x, tjs_int &y)
{
	// vC}C}l[W̃vC}C̃TCY𓾂
	tjs_int pl_w, pl_h;
	manager->GetPrimaryLayerSize(pl_w, pl_h);
	x = pl_w ? (x * destWidth  / pl_w) : 0;
	y = pl_h ? (y * destHeight / pl_h) : 0;
}

/**
 * DeviceWʂ̍W̕ϊs
 * @param		x		Xʒu
 * @param		y		Yʒu
 * @note		x, y  DestRect (0,0) _ƂWƂēnƌȂ
 */
void
IrrlichtDrawDevice::transformTo(tjs_int &x, tjs_int &y)
{
	x = destWidth  ? (x * width / destWidth) : 0;
	y = destHeight ? (y * height / destHeight) : 0;
}

/** WʁDevice̍W̕ϊs
 * @param		x		Xʒu
 * @param		y		Yʒu
 * @note		x, y  C (0,0) _ƂWƂēnƌȂ
 */
void
IrrlichtDrawDevice::transformFrom(tjs_int &x, tjs_int &y)
{
	// vC}C}l[W̃vC}C̃TCY𓾂
	x = width ? (x * destWidth  / width) : 0;
	y = height ? (y * destHeight / height) : 0;
}

/**
 * ŗLXV
 */
void
IrrlichtDrawDevice::update(irr::video::IVideoDriver *driver)
{
	// ʃC}l[W̕`
	for (std::vector<iTVPLayerManager *>::iterator i = Managers.begin(); i != Managers.end(); i++) {
		LayerManagerInfo *info = (LayerManagerInfo*)(*i)->GetDrawDeviceData();
		if (info) {
			info->draw(driver, screenRect);
		}
	}
}


/**
 * C}l[W̓o^
 * @param manager C}l[W
 */
void TJS_INTF_METHOD
IrrlichtDrawDevice::AddLayerManager(iTVPLayerManager * manager)
{
	int id = (int)Managers.size();
	tTVPDrawDevice::AddLayerManager(manager);
	LayerManagerInfo *info = new LayerManagerInfo(id, defaultVisible);
	manager->SetDrawDeviceData((void*)info);
}

/**
 * C}l[W̍폜
 * @param manager C}l[W
 */
void TJS_INTF_METHOD
IrrlichtDrawDevice::RemoveLayerManager(iTVPLayerManager * manager)
{
	LayerManagerInfo *info = (LayerManagerInfo*)manager->GetDrawDeviceData();
	if (info != NULL) {
		manager->SetDrawDeviceData(NULL);
		delete info;
	}
	tTVPDrawDevice::RemoveLayerManager(manager);
}

/***
 * EChE̎w
 * @param wnd EChEnh
 */
void TJS_INTF_METHOD
IrrlichtDrawDevice::SetTargetWindow(HWND wnd, bool is_main)
{
	stop();
	detach();
	if (wnd != NULL) {
		attach(wnd, screenWidth, screenHeight);
		start();
	}
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::SetDestRectangle(const tTVPRect &dest)
{
	destRect.setLeft(dest.left);
	destRect.setTop(dest.top);
	destRect.setWidth((destWidth = dest.get_width()));
	destRect.setHeight((destHeight = dest.get_height()));
	if (device) {
		IVideoDriver *driver = device->getVideoDriver();
		if (driver) {
			tjs_int w, h;
			if (zoomMode) {
				w = width;
				h = height;
			} else {
				w = destWidth;
				h = destHeight;
			}
			if (screenWidth != w ||	screenHeight != h) {
				screenWidth = w;
				screenHeight = h;
				screenRect = rect<s32>(0,0,screenWidth,screenHeight);
				driver->OnResize(dimension2d<s32>(w, h));
			}
		}
	}
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::GetSrcSize(tjs_int &w, tjs_int &h)
{
	w = width;
	h = height;
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::NotifyLayerResize(iTVPLayerManager * manager)
{
	LayerManagerInfo *info = (LayerManagerInfo*)manager->GetDrawDeviceData();
	if (info != NULL) {
		info->free();
		if (device) {
			IVideoDriver *driver = device->getVideoDriver();
			if (driver) {
				info->alloc(manager, driver);
			}
		}
	}
}

// -------------------------------------------------------------------------------------
// ̓Cxgp
// -------------------------------------------------------------------------------------

void TJS_INTF_METHOD
IrrlichtDrawDevice::OnMouseDown(tjs_int x, tjs_int y, tTVPMouseButton mb, tjs_uint32 flags)
{
	// Irrlicht ɑ
	if (device) {
		tjs_int dx = x;
		tjs_int dy = y;
		transformToIrrlicht(dx, dy);
		SEvent ev;
		ev.EventType = EET_MOUSE_INPUT_EVENT;
		ev.MouseInput.X = dx;
		ev.MouseInput.Y = dy;
		ev.MouseInput.Wheel = 0;
		switch ((mb & 0xff)) {
		case mbLeft:
			ev.MouseInput.Event = EMIE_LMOUSE_PRESSED_DOWN;
			break;
		case mbMiddle:
			ev.MouseInput.Event = EMIE_MMOUSE_PRESSED_DOWN;
			break;
		case mbRight:
			ev.MouseInput.Event = EMIE_RMOUSE_PRESSED_DOWN;
			break;
		}
		postEvent(ev);
	}
	// gg̃vC}Cɑ
	iTVPLayerManager * manager = GetLayerManagerAt(PrimaryLayerManagerIndex);
	if (manager) {
		transformToManager(manager, x, y);
		manager->NotifyMouseDown(x, y, mb, flags);
	}
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::OnMouseUp(tjs_int x, tjs_int y, tTVPMouseButton mb, tjs_uint32 flags)
{
	// Irrlicht ɑ
	if (device) {
		tjs_int dx = x;
		tjs_int dy = y;
		transformToIrrlicht(dx, dy);
		SEvent ev;
		ev.EventType = EET_MOUSE_INPUT_EVENT;
		ev.MouseInput.X = dx;
		ev.MouseInput.Y = dy;
		ev.MouseInput.Wheel = 0;
		switch ((mb & 0xff)) {
		case mbLeft:
			ev.MouseInput.Event = EMIE_LMOUSE_LEFT_UP;
			break;
		case mbMiddle:
			ev.MouseInput.Event = EMIE_MMOUSE_LEFT_UP;
			break;
		case mbRight:
			ev.MouseInput.Event = EMIE_RMOUSE_LEFT_UP;
			break;
		}
		postEvent(ev);
	}
	// gg̃vC}Cɑ
	iTVPLayerManager * manager = GetLayerManagerAt(PrimaryLayerManagerIndex);
	if (manager) {
		transformToManager(manager, x, y);
		manager->NotifyMouseUp(x, y, mb, flags);
	}
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::OnMouseMove(tjs_int x, tjs_int y, tjs_uint32 flags)
{
	// Irrlicht ɑ
	if (device) {
		tjs_int dx = x;
		tjs_int dy = y;
		transformToIrrlicht(dx, dy);
		SEvent ev;
		ev.EventType = EET_MOUSE_INPUT_EVENT;
		ev.MouseInput.X = dx;
		ev.MouseInput.Y = dy;
		ev.MouseInput.Wheel = 0;
		ev.MouseInput.Event = EMIE_MOUSE_MOVED;
		postEvent(ev);
	}
	// gg̃vC}Cɑ
	iTVPLayerManager * manager = GetLayerManagerAt(PrimaryLayerManagerIndex);
	if (manager) {
		transformToManager(manager, x, y);
		manager->NotifyMouseMove(x, y, flags);
	}
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::OnMouseWheel(tjs_uint32 shift, tjs_int delta, tjs_int x, tjs_int y)
{
	// Irrlicht ɑ
	if (device) {
		tjs_int dx = x;
		tjs_int dy = y;
		transformToIrrlicht(dx, dy);
		SEvent ev;
		ev.EventType = EET_MOUSE_INPUT_EVENT;
		ev.MouseInput.X = dx;
		ev.MouseInput.Y = dy;
		ev.MouseInput.Wheel = (f32)delta;
		ev.MouseInput.Event = EMIE_MOUSE_WHEEL;
		postEvent(ev);
	}
	// gg̃vC}Cɑ
	iTVPLayerManager * manager = GetLayerManagerAt(PrimaryLayerManagerIndex);
	if (manager) {
		transformToManager(manager, x, y);
		manager->NotifyMouseWheel(shift, delta, x, y);
	}
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::GetCursorPos(iTVPLayerManager * manager, tjs_int &x, tjs_int &y)
{
	Window->GetCursorPos(x, y);
	transformToManager(manager, x, y);
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::SetCursorPos(iTVPLayerManager * manager, tjs_int x, tjs_int y)
{
	transformFromManager(manager, x, y);
	Window->SetCursorPos(x, y);
}

void TJS_INTF_METHOD
IrrlichtDrawDevice::RequestInvalidation(const tTVPRect & rect)
{
	for (std::vector<iTVPLayerManager *>::iterator i = Managers.begin(); i != Managers.end(); i++) {
		iTVPLayerManager *manager = *i;
		LayerManagerInfo *info = (LayerManagerInfo*)manager->GetDrawDeviceData();
		if (info && info->visible) {
			tjs_int l = rect.left, t = rect.top, r = rect.right, b = rect.bottom;
			transformToManager(manager, l, t);
			transformToManager(manager, r, b);
			r ++; // 덷̋z({͂ƌɂȂƂȂȂꂪɂȂ邱Ƃ͂Ȃ)
			b ++;
			manager->RequestInvalidation(tTVPRect(l, t, r, b));
		}
	}
}


// -------------------------------------------------------------------------------------
// ĕ`揈p
// -------------------------------------------------------------------------------------

void
IrrlichtDrawDevice::Show()
{
	show(&destRect);
}

// -------------------------------------------------------------------------------------
// LayerManager̉摜킽
// -------------------------------------------------------------------------------------

/**
 * rbg}bvRs[Jn
 */
void TJS_INTF_METHOD
IrrlichtDrawDevice::StartBitmapCompletion(iTVPLayerManager * manager)
{
	LayerManagerInfo *info = (LayerManagerInfo*)manager->GetDrawDeviceData();
	if (info) {
		info->lock();
	}
}

/**
 * rbg}bvRs[
 */
void TJS_INTF_METHOD
IrrlichtDrawDevice::NotifyBitmapCompleted(iTVPLayerManager * manager,
	tjs_int x, tjs_int y, const void * bits, const BITMAPINFO * bitmapinfo,
	const tTVPRect &cliprect, tTVPLayerType type, tjs_int opacity)
{
	LayerManagerInfo *info = (LayerManagerInfo*)manager->GetDrawDeviceData();
	if (info) {
		info->copy(x, y, bits, bitmapinfo, cliprect, type, opacity);
	}
}

/**
 * rbg}bvRs[I
 */
void TJS_INTF_METHOD
IrrlichtDrawDevice::EndBitmapCompletion(iTVPLayerManager * manager)
{
	LayerManagerInfo *info = (LayerManagerInfo*)manager->GetDrawDeviceData();
	if (info) {
		info->unlock();
	}
}

/**
 * vC}C̕\Ԃ̎w
 * @param id vC}C̓o^ID
 * @param visible \
 */
void
IrrlichtDrawDevice::setVisible(int id, bool visible)
{
	if (id >= 0 && id < (int)Managers.size()) {
		LayerManagerInfo *info = (LayerManagerInfo*)Managers[id]->GetDrawDeviceData();
		if (info) {
			info->visible = visible;
		}
	}
}

/**
 * vC}C̕\Ԃ̎w
 * @param id vC}C̓o^ID
 * @return visible \
 */
bool
IrrlichtDrawDevice::getVisible(int id)
{
	if (id >= 0 && id < (int)Managers.size()) {
		LayerManagerInfo *info = (LayerManagerInfo*)Managers[id]->GetDrawDeviceData();
		if (info) {
			return info->visible;
		}
	}
	return false;
}
